/* -*- mode: c; c-basic-offset: 4; -*- */
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>

#include "channel.h"
#include "daqc.h"
#include "daqc_response.h"
#include "daqc_internal.h"

/*--------------------------------------------------------------------
 *
 *    Select includes depend on posix versions
 *--------------------------------------------------------------------*/
#if _POSIX_C_SOURCE < 200100L
#include <sys/time.h>
#include <sys/types.h>
#else
#include <sys/select.h>
#endif

/*  Swap the specified data in 
 */
int
swap_data(int gran, int nBytes, char* data) {
    if (gran<2 || ntohs(1) == 1) return nBytes/gran;
    int i, n=0;
    if (gran == 2 && !((long)data & 1)) {
	short* p = (short*)data;
	for (i=0; i<nBytes; i+=gran) {
	    *p = ntohs(*p);
	    p++; n++;
	}
    } else if (gran == 4 && !((long)data & 3)) {
	int* p = (int*)data;
	for (i=0; i<nBytes; i+=gran) {
	    *p = ntohl(*p);
	    p++; n++;
	}
    } else {
	for (i=0; i<nBytes; i += gran) {	    
	    char* head = data + i;
	    char* tail = head + gran;
	    while (head < tail) {
		char t  = *head;
		*head++ = *(--tail);
		*tail   = t;
	    }
	    n++;
	}
    }
    return n;
}

/* convert nds1 group number to nds2 type code.
 */
enum chantype
cvt_group_chantype(int group_num) {
    switch (group_num) {
    case 1000:
	return cMTrend;
    case 0:
	return cOnline;
    default:
	break;
    }
    return cUnknown;
}

/**  Convert a null terminated hex string to a long integer.
 *  \param str Pointer to hexidecimal string.
 *  \return Value of hex string or -1 if syntax error.
 */
long
dca_strtol (char *str) {
    char *ptr;
    long res = strtol (str, &ptr, 16);
    if (*ptr) return -1;
    return res;
}

/*  Wait for data on the specified port. 
 *  return values:
 *   1 Data available
 *   0 Timeout.
 *  -1 Error
 */
int
_daq_wait_data(int fd, int wt) {
    struct timeval tspec, *tptr;
    tspec.tv_sec  = wt;
    tspec.tv_usec = 0;
    tptr = &tspec;

    fd_set readfds;
    FD_ZERO(&readfds);
    FD_SET(fd, &readfds);

    int rc = -1;
    while (rc < 0) {
	rc = select(fd+1, &readfds, 0, 0, tptr);
	if (rc < 0 && errno != EINTR) break;
    }
    if (!rc) {
	fprintf(stderr, "Timeout in _daq_wait_data\n");
    } else if (rc < 0) {
	perror("Error in _daq_wait_data()");
    }
    return rc;
}

/* 
 *   Read `numb' bytes.
 *   Returns number of bytes read or 0 on the error or EOF
 */
int
read_bytes (int fd, char *cptr, int numb) {
    int oread = 0;
    while (oread < numb) {
	int bread = read (fd, cptr+oread, numb-oread);
	if (bread > 0) {
	    oread += bread;
	} else if (!bread) {
	    break;
	} else if (errno != EAGAIN) {
#ifdef DAQC_ACCESS_VERBOSE_ERRORS
	    perror("read() error");
#endif
	    return 0;
        }
    }
    /* printf("read(%i,...,%i) = %i\n", fd, numb,oread); */
    return oread;
}

/* Read server response -- 4 bytes */
/* Convert the response to integer */
/* Return -1 for invalid response */
long
read_server_response_wait(int fd, int wt) {
    if (wt > 0 && _daq_wait_data(fd, wt) <= 0) {
	perror("read_server_response_wait: Error waiting for data");
	return -1;
    }
    return read_server_response(fd);
}

long 
read_server_response (int fd) { 
#define response_length 4
    char buf [response_length + 1];
    int rc = read_bytes(fd, buf, response_length);
    if (rc != response_length) {
	if (rc < 0) perror("read_server_response: Error in read");
	else fprintf(stderr, "read_server_response: Wrong length read (%i)\n",
		     rc);
	return -1;
    }
    buf [response_length] = '\000';
    return dca_strtol (buf);
#undef response_length
}

/*  read_float(int fd, float*)
 *
 *  read a float from the indicated fd and swap bytes from network
 *  byte ordering (big-endian) to host byte ordering.
 */
int
read_float(int fd, float* data) {
    union {uint32_t i; float f;} d;
    if (read_uint4(fd, &d.i)) return DAQD_ERROR;
    *data = d.f;
    return DAQD_OK;
}

/*  Read a 4-byte `long' integer. The byte ordering is swapped on little-
 *  endian machines.
 */
int
read_uint4 (int fd, uint4_type* data) {
    uint4_type buf;
    if (read_bytes(fd, (char*)&buf, sizeof(buf)) != sizeof(buf)) 
	return DAQD_ERROR;
    *data = ntohl(buf);
    return DAQD_OK;
}

/*
 *  Read an NDS2 byte-counted string from the specified fd. The string is 
 *  transferred as an integer byte-count followed by text.
 */
int
_daq_read_string(int fd, uint4_type maxlen, char* buf) {
    uint4_type len = 0;
    if (read_uint4(fd, &len)) return -1;
    if (len == 0) return len;
    int rc;
    if (len > maxlen) rc = read_bytes(fd, buf, maxlen);
    else              rc = read_bytes(fd, buf, len);
    if (rc < 0) return -1;
    if (rc < maxlen) buf[rc] = 0;
    return len;
}

int
null_term(char* p, int len) {
    int i;
    for (i=len-1; i; --i) {
	if (!isspace(p[i-1])) {
	    p[i] = 0;
	    return i;
	}
    }
    *p = 0;
    return 0;
}

/** read data from string pointed to by *p. skip past any leading spaces.
  * copy all data to out until a space or a null is found or out is full. 
  * return the number of characters copied and increment *p as appropriate.
  */
int 
_daq_get_string(const char** p, char* out, int lmax) {
    int i;
    const char* in = *p;
    while (*in == ' ') in++;
    *p = in;
    for (i=0; i<lmax; i++) {
	if (!*in || *in == ' ') break;
	out[i] = *in++;
    }
    if (i < lmax) out[i] = 0;
    *p += i;
    return i;
}

/*  _daq_cvt_string(long, char*)
 *
 *  convert the specified long int to a string
 */
int
_daq_cvt_string(long in, char* out) {
    int n = 0;
    if (in >= 10) n = _daq_cvt_string(in/10, out);
    out[n] = '0' + (in%10);
    return n+1;
}

/*
 *  Convert the specified host name and port number into an ip address
 */
int
daq_set_server(daq_t *daq, const char* host, int port) {
#if defined(__linux__) || defined(__APPLE__)
    extern int h_errno;
    struct hostent* hentp = gethostbyname (host);
    if (!hentp) {
	fprintf (stderr, "Can't find hostname `%s'\n", host);
#ifdef DAQC_ACCESS_VERBOSE_ERRORS
	perror("gethostbyname() error");
#endif
	return DAQD_INVALID_IP_ADDRESS;
    }
    memcpy(&daq -> srvr_addr.sin_addr.s_addr, *hentp -> h_addr_list, 
	   sizeof (daq -> srvr_addr.sin_addr.s_addr));

#else
#define HOST_BUF_LENGTH 2048
    struct hostent hent;
    char buf [HOST_BUF_LENGTH];
    int gherr;
    if (! gethostbyname_r (host, &hent, buf, HOST_BUF_LENGTH, &gherr)) {
	fprintf (stderr, "Can't find hostname `%s'\n", host);
#ifdef DAQC_ACCESS_VERBOSE_ERRORS
	perror("gethostbyname_r() error");
#endif
	return DAQD_INVALID_IP_ADDRESS;
    }

    memcpy(&daq->srvr_addr.sin_addr.s_addr, *hent.h_addr_list, 
	   sizeof(daq->srvr_addr.sin_addr.s_addr));
#endif

    daq -> srvr_addr.sin_family = AF_INET;
    /*  daq -> srvr_addr.sin_addr.s_addr = inet_addr (ip); */
    daq -> srvr_addr.sin_port = htons (port? port: DAQD_PORT);
    return DAQD_OK;
}

/*--------------------------------------------------------------------------*
 *                                                                          *
 *         Obsolete functions                                               *
 *                                                                          *
 *--------------------------------------------------------------------------*/

/*  daq_recv_channel_groups()
 *
 *  Get channel groups description
 *  Returns zero on success.
 *  Sets `*num_channel_groups_received' to the actual number of 
 *  configured channel groups.
 */
int
daq_recv_channel_groups (daq_t *daq, daq_channel_group_t *group,
			 int num_groups, int *num_channel_groups_received)
{
    int i;
    int resp;
    int groups;

    if ((resp = daq_send (daq, "status channel-groups;")))
	return resp;

    /* Read the number of channels */
    if ((groups = read_server_response (daq -> sockfd)) <= 0) {
	fprintf (stderr, "couldn't determine the number of channel groups\n");
	return DAQD_ERROR;
    }

    *num_channel_groups_received = groups;
    if (num_groups < groups)
	groups = num_groups;

    for (i = 0; i < groups; i++) {
	int len=read_bytes(daq->sockfd, group[i].name, MAX_CHANNEL_NAME_LENGTH);
	if (len != MAX_CHANNEL_NAME_LENGTH)
	    return DAQD_ERROR;
	null_term(group [i].name, len);

	group [i].group_num = read_server_response (daq -> sockfd);
	if (group [i].group_num < 0) return DAQD_ERROR;
    }
    return 0;
}
