/* -*- mode: c; c-basic-offset: 4; -*- */
/*******************************************************
  Matlab mex wrapper for NDS.
******************************************************/

#include <string.h>
#include "daqc.h"
#include "nds_mex_utils.h"

/* for pre v7.3 compatibility */
#ifndef mwSize
#define mwSize int
#endif

mxArray*
mlwrapper_chanlist_to_mxarray(int *err,daq_channel_t* channels,int num_channels);
int
mlwrapper_get_channel_list(mxArray **, char*, chantype_t, time_t);

/*   NDS_GetChannels function */
/*   ------------------------ */
/*   This function returns a list of available channels */
/*   The syntax is:                                     */
/*   NDS_GetChannels("server:port");                    */
void 
mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    char hpbuf[MAX_HOST_LENGTH];
    int buflen;
    int err;

    /*** Input Validation ***/
    if(nlhs != 1) {
	mexErrMsgTxt("This function only returns one paramter.");
    }

    if(nrhs < 1 || nrhs > 3) {
	mexErrMsgTxt("This function takes 1-3 arguments.");
    }

    /*--- Get the channel type if a second argument was specified */
    chantype_t ctype = cUnknown;
    if (nrhs > 1) {
	if (mxIsChar(prhs[1]) != 1) {
	    mexErrMsgTxt("Channel type must be speciified as a string");
	} else {
	    buflen = (mxGetM(prhs[1]) * mxGetN(prhs[1]));
	    if(buflen >= MAX_HOST_LENGTH) {
		mexErrMsgTxt("Channel type string too long.");
	    }
	    mxGetString(prhs[1], hpbuf, buflen + 1);
	    hpbuf[buflen] = 0;
	    ctype = cvt_str_chantype(hpbuf);
	}
    }

    /*--- Get the host name                                      */
    if(mxIsChar(prhs[0]) != 1) {
	mexErrMsgTxt("Argument 1 must be a string of the format 'host:port'");
    }
    buflen = (mxGetM(prhs[0]) * mxGetN(prhs[0]));
    if(buflen >= MAX_HOST_LENGTH) {
	mexErrMsgTxt("host:port string too long.");
    }
    mxGetString(prhs[0], hpbuf, buflen + 1);

    /*--- Get the gps time if it was specified                  */
    time_t gps = 0;
    if (nrhs > 2) {
	if((mxGetM(prhs[2]) * mxGetN(prhs[2])) != 1) {
	    mexErrMsgTxt("Start time must be a scalar.");
	}
	gps = mxGetScalar(prhs[2]);
    }

    err = mlwrapper_get_channel_list(plhs, hpbuf, ctype, gps);

    if(err) {
	mexErrMsgTxt("Failed to get channel list.");
    }
}

int
mlwrapper_get_channel_list(mxArray **plhs, char* hpbuf, 
			   chantype_t chant, time_t gps) {
    /* Initialization processing */
    daq_startup();

    /* Get the host ip and port number */
    char* host = NULL;
    short port = 31200;
    parse_host_name(hpbuf, &host, &port);
    if (!host) {
	mexWarnMsgTxt("Failed to parse host:port string.");
	return 11;
    }

    /* Open up a client connection */
    daq_t daq;
    int err = daq_connect(&daq, host, port, nds_try);
    /* printf("daq_connect: return code %i\n", err); */
    free(host);
    if (err) {
	mexWarnMsgTxt("Failed to connect to nds2.");
	return 13;
    }

    /*---  Get the number of channels */
    int num_alloc    = 0;
    daq_channel_t* channels = NULL;
    err = daq_recv_channel_list(&daq, channels, 0, &num_alloc, gps, chant);
    if (err) {
	daq_recv_shutdown(&daq);
	mexWarnMsgTxt("get_channel_list() failed.");
	return 54;
    }

    /*---  Read in the channel list */
    int num_channels = 0;
    if (num_alloc > 0) {
	channels = malloc(num_alloc * sizeof(daq_channel_t));
	if (!channels) {
	    mexWarnMsgTxt("Channel list malloc() failed.");
            return 54;
	}
	err = daq_recv_channel_list(&daq, channels, num_alloc, 
				    &num_channels, gps, chant);
	if (err) {
	    free(channels);
	    mexWarnMsgTxt("get_channel_list() failed.");
	    return 54;
	} else if (!num_channels) {
	    free(channels);
	    channels = NULL;
	}
    }

    /*---  Close the connection */
    daq_disconnect(&daq);
    daq_recv_shutdown(&daq);

    /*---  Convert the channel list to a cell-array */
    plhs[0] = mlwrapper_chanlist_to_mxarray(&err, channels, num_channels);
    if (err) {
	mexWarnMsgTxt("mlwrapper_chanlist_to_mxarray() failed.");
	if (channels != NULL) {
	    free(channels);
	}
	return 83;
    }

    if (channels != NULL) {
	free(channels);
    }

    return 0;
}

mxArray*
mlwrapper_chanlist_to_mxarray(int *err,daq_channel_t* channels,int num_channels)
{
    /* Retrieves the channel list from the NDS server,
       and converts it into a mxArray of Matlab structs */
    mxArray *ml_list;
    mwSize dims[1];
    char* fields[] = {"name", "chan_type", "rate", "data_type", "signal_gain", 
		      "signal_offset", "signal_slope", "signal_units"};

    dims[0] = (mwSize) num_channels;

    ml_list = mxCreateStructArray(1,dims,8,(const char**)fields);
    if(ml_list == NULL) {
	*err = 2;
	return NULL;
    }

    /*---  Bail out if no channels are available. */
    if (num_channels == 0) {
	return ml_list;
    }

    int field_name = mxGetFieldNumber(ml_list,"name");
    int field_chan_type = mxGetFieldNumber(ml_list,"chan_type");
    int field_rate = mxGetFieldNumber(ml_list,"rate");
    int field_data_type = mxGetFieldNumber(ml_list,"data_type");
    int field_signal_gain = mxGetFieldNumber(ml_list,"signal_gain");
    int field_signal_offset = mxGetFieldNumber(ml_list,"signal_offset");
    int field_signal_slope = mxGetFieldNumber(ml_list,"signal_slope");
    int field_signal_units = mxGetFieldNumber(ml_list,"signal_units");

    int i;
    for(i=0; i < num_channels; i++) {
	/* Channel Name */
	mxArray* temp = mxCreateString(channels[i].name);
	if(temp == NULL) {
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}
	mxSetFieldByNumber(ml_list, i, field_name, temp);
	
	/* Channel type (formerly group number) */
	temp = mxCreateString(cvt_chantype_str(channels[i].type));
	if(temp == NULL) {
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}
	mxSetFieldByNumber(ml_list, i, field_chan_type, temp);

	/* Sample Rate */
	if (put_mxarray_int(ml_list, field_rate, i, channels[i].rate)){
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}

	/* Data Type */
	temp = mxCreateString(data_type_name(channels[i].data_type));
	if(temp == NULL) {
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}
	mxSetFieldByNumber(ml_list, i, field_data_type, temp);

	/* Signal Gain */
	if (put_mxarray_float(ml_list, field_signal_gain, i, 
			      channels[i].s.signal_gain)){
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}

	/* Signal Offset */
	if (put_mxarray_float(ml_list, field_signal_offset, i, 
			      channels[i].s.signal_offset)){
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}

	/* Signal Slope */
	if (put_mxarray_float(ml_list, field_signal_slope, i, 
			      channels[i].s.signal_slope)){
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}

	/* Signal Units */
	temp = mxCreateString(channels[i].s.signal_units);
	if(temp == NULL) {
	    mxDestroyArray(ml_list);
	    *err = 3;
	    return NULL;
	}
	mxSetFieldByNumber(ml_list,i,field_signal_units,temp);
    }
    return ml_list;
}
